/*
 * str-3143-threads.cxx
 */


#include <cstdio>

#ifdef _WIN32
#  include <windows.h>
#  include <process.h>    /* _beginthread, _endthread */
#  define THREAD HANDLE
#  define wait_here(x) Sleep(x)
#else /* for *nix and OSX */
#  include <pthread.h>
#  include <sched.h> // maybe necessary for sched_yield ?
#  include <unistd.h>
#  define THREAD pthread_t
#  define wait_here(x) usleep((x)*1000)
#endif

#include <FL/Fl.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Text_Display.H>
#include <FL/Fl_Window.H>

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

class ThreadApp
{
private:
    Fl_Button* startButton;
    Fl_Button* nudgeButton;
    Fl_Text_Buffer* msgBuffer;

    THREAD workThread;
    unsigned int count;
public:
    ThreadApp();

    Fl_Window* mainWin;

    static void awakeMsg(void* data);
    static void awakeEnd(void* data);
    static void startCalled(Fl_Widget* w, void* data);
    static void nudgeCalled(Fl_Widget* w, void* data);
    static void exiting(Fl_Widget* w, void* data);

#ifdef _WIN32
    static unsigned WINAPI workThreadCallback(LPVOID data);
#else /* for *nix and OSX */
    static void *workThreadCallback(void *data);
#endif

    void workThreadFunc();
};

ThreadApp::ThreadApp()
    : mainWin(NULL),
      workThread(NULL)
{
    mainWin = new Fl_Window(600, 400, "ThreadApp");
    Fl_Text_Display* msgView = new Fl_Text_Display(10, 10, 580, 350);
    msgBuffer = new Fl_Text_Buffer();
    msgView->buffer(msgBuffer);
    startButton = new Fl_Button(10, 360, 70, 30, "START");
    startButton->callback(&ThreadApp::startCalled, this);

    nudgeButton = new Fl_Button(90, 360, 70, 30, "Nudge");
    nudgeButton->callback(&ThreadApp::nudgeCalled, this);

    mainWin->callback(&ThreadApp::exiting, this);
    mainWin->end();
    mainWin->resizable(msgView);
}

void ThreadApp::awakeMsg(void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    char buf[64];
    snprintf(buf, 64, "Message: %u\n", app->count++);
    app->msgBuffer->append(buf);
}

void ThreadApp::awakeEnd(void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->msgBuffer->append("Finish");
    app->workThread = NULL;
    app->startButton->activate(); // enable
}

void ThreadApp::startCalled(Fl_Widget* w, void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    if (app->workThread != NULL)
        return; // ignore if ran
    // starting thread
    app->msgBuffer->remove(0, app->msgBuffer->length());
    app->count = 0;
    app->startButton->deactivate();

#ifdef _WIN32
    app->workThread = (THREAD)_beginthreadex(NULL, 0, &ThreadApp::workThreadCallback, app, 0, NULL);
#else /* for *nix and OSX */
  pthread_t thread;
  int res = pthread_create(&thread, NULL, &ThreadApp::workThreadCallback, app);
  if (res == 0)
    app->workThread = thread;
  else
    app->workThread = NULL;
#endif
}

void ThreadApp::nudgeCalled(Fl_Widget* w, void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->mainWin->redraw();
    puts("nudge!"); fflush(stdout);
    if (app->workThread != NULL) Fl::awake();
}

#ifdef _WIN32
unsigned WINAPI ThreadApp::workThreadCallback(LPVOID data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->workThreadFunc();
    return 0;
}
#else /* for *nix and OSX */
void *ThreadApp::workThreadCallback(void *data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    app->workThreadFunc();
    return NULL;
}
#endif

void ThreadApp::workThreadFunc()
{
    for (int i = 0; i < 5; i++)
    {
       int res = Fl::awake(&ThreadApp::awakeMsg, this);

        wait_here(1000);
        printf("tick %d (%d)\n", i, res); fflush(stdout);
    }
    // end of thread
    int res = Fl::awake(&ThreadApp::awakeEnd, this);
    printf("out (%d)\n", res); fflush(stdout);
}

void ThreadApp::exiting(Fl_Widget* w, void* data)
{
    ThreadApp* app = reinterpret_cast<ThreadApp*>(data);
    if (app->workThread == NULL)
        w->hide();
}

static void heartbeat(void *)
{
    static int checkit = 10;
    printf("Heartbeat: %d\n", checkit); fflush(stdout);
    Fl::repeat_timeout(2.0, heartbeat);

    --checkit;
    if (checkit < 0)
    {
        checkit = 10;
        //Fl::awake();
    }
}

int main(int argc, char** argv)
{
    Fl::lock();

    ThreadApp *app = new ThreadApp();
    app->mainWin->show(argc, argv);

    Fl::add_timeout(0.5, heartbeat);

    return Fl::run();
}

/* End of file */
